@@ -21,7 +21,7 @@ struct O5mStringPair {
     std::string s1;
     std::string s2;
 
-    inline bool operator==(const O5mStringPair &other) const {
+    [[nodiscard]] inline bool operator==(const O5mStringPair &other) const {
         return s1 == other.s1 && s2 == other.s2;
     }
 };
@@ -31,7 +31,7 @@ template<>
 struct std::hash<OSM::O5mStringPair>
 {
 
-    std::size_t operator()(const OSM::O5mStringPair &p) const noexcept
+    [[nodiscard]] std::size_t operator()(const OSM::O5mStringPair &p) const noexcept
     {
         std::size_t h1 = std::hash<std::string>{}(p.s1);
         std::size_t h2 = std::hash<std::string>{}(p.s2);
